from abc import ABC, abstractmethod
from telegrinder.api.abc import ABCAPI


class ABCDispatch(ABC):
    @abstractmethod
    def feed(self, event: dict, api: ABCAPI) -> bool:
        pass

    @abstractmethod
    def load(self, external: "ABCDispatch"):
        pass



import typing
from typing import List, Optional
from dataclasses import dataclass
from abc import ABC, abstractmethod
from telegrinder.types.objects import InlineKeyboardMarkup, ReplyKeyboardMarkup

from .buttons import Button, InlineButton


AnyMarkup = typing.Union[InlineKeyboardMarkup, ReplyKeyboardMarkup]


@dataclass()
class KeyboardModel:
    resize_keyboard: bool
    one_time_keyboard: bool
    selective: bool
    keyboard: List[List[dict]]


class ABCMarkup(ABC, KeyboardModel):
    BUTTON: typing.Any

    def __init__(
        self,
        resize_keyboard: bool = True,
        one_time_keyboard: bool = False,
        selective: Optional[bool] = None,
    ):
        self.keyboard = [[]]
        self.resize_keyboard = resize_keyboard
        self.one_time_keyboard = one_time_keyboard
        self.selective = selective

    @abstractmethod
    def add(self, button) -> "ABCMarkup":
        pass

    @abstractmethod
    def row(self) -> "ABCMarkup":
        pass

    @abstractmethod
    def dict(self) -> dict:
        pass

    @abstractmethod
    def get_markup(self) -> AnyMarkup:
        pass

    @classmethod
    def empty(cls) -> AnyMarkup:
        return cls().get_markup()

    def format(self, **format_data: typing.Dict[str, str]) -> "ABCMarkup":
        copy_keyboard = self.__class__()
        for row in self.keyboard:
            for button in row:
                copy_button = button.copy()
                copy_button["text"] = copy_button["text"].format(**format_data)
                copy_keyboard.add(self.BUTTON(**copy_button))
            copy_keyboard.row()
        return copy_keyboard


class Keyboard(ABCMarkup):
    BUTTON = Button

    def row(self) -> "Keyboard":
        if len(self.keyboard) and not len(self.keyboard[-1]):
            raise RuntimeError("Last row is empty!")

        self.keyboard.append([])
        return self

    def add(self, button: Button) -> "Keyboard":
        if not len(self.keyboard):
            self.row()

        self.keyboard[-1].append(button.get_data())
        return self

    def dict(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def get_markup(self) -> ReplyKeyboardMarkup:
        return ReplyKeyboardMarkup(**self.dict())


class InlineKeyboard(ABCMarkup):
    BUTTON = InlineButton

    def row(self) -> "InlineKeyboard":
        if len(self.keyboard) and not len(self.keyboard[-1]):
            raise RuntimeError("Last row is empty!")

        self.keyboard.append([])
        return self

    def add(self, button: InlineButton) -> "InlineKeyboard":
        if not len(self.keyboard):
            self.row()

        self.keyboard[-1].append(button.get_data())
        return self

    def dict(self) -> dict:
        return dict(inline_keyboard=self.keyboard)

    def get_markup(self) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup(**self.dict())



import typing
from .objects import *
from telegrinder.tools import Result
from telegrinder.api.error import APIError

if typing.TYPE_CHECKING:
    from telegrinder.api.abc import ABCAPI

X = typing.TypeVar("X")


class APIMethods:
    def __init__(self, api: "ABCAPI"):
        self.api = api

    @staticmethod
    def get_params(loc: dict) -> dict:
        n = {k: v for k, v in loc.items() if k not in ("self", "other") and v is not None}
        n.update(loc['other'])
        return n

    @staticmethod
    def get_response(r: dict) -> dict:
        if 'json' in r: r['json_'] = r['json']
        return r

    async def get_updates(
        self,
        offset: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        timeout: typing.Optional[int] = None,
        allowed_updates: typing.Optional[typing.List[str]] = None,
        **other
    ) -> Result[typing.List['Update'], APIError]:
        result = await self.api.request("getUpdates", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=[Update(**self.get_response(e)) for e in u])
        return Result(False, error=result.error)

    async def set_webhook(
        self,
        url: typing.Optional[str] = None,
        certificate: typing.Optional['InputFile'] = None,
        ip_address: typing.Optional[str] = None,
        max_connections: typing.Optional[int] = None,
        allowed_updates: typing.Optional[typing.List[str]] = None,
        drop_pending_updates: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setWebhook", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def delete_webhook(
        self,
        drop_pending_updates: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("deleteWebhook", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_webhook_info(
        self,
        **other
    ) -> Result['WebhookInfo', APIError]:
        result = await self.api.request("getWebhookInfo", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=WebhookInfo(**self.get_response(u)))
        return Result(False, error=result.error)

    async def get_me(
        self,
        **other
    ) -> Result['User', APIError]:
        result = await self.api.request("getMe", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=User(**self.get_response(u)))
        return Result(False, error=result.error)

    async def log_out(
        self,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("logOut", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def close(
        self,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("close", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def send_message(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        text: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        entities: typing.Optional[typing.List['MessageEntity']] = None,
        disable_web_page_preview: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendMessage", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def forward_message(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        from_chat_id: typing.Optional[typing.Union[int, str]] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        message_id: typing.Optional[int] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("forwardMessage", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def copy_message(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        from_chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['MessageId', APIError]:
        result = await self.api.request("copyMessage", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=MessageId(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_photo(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        photo: typing.Optional[typing.Union['InputFile', str]] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendPhoto", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_audio(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        audio: typing.Optional[typing.Union['InputFile', str]] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        duration: typing.Optional[int] = None,
        performer: typing.Optional[str] = None,
        title: typing.Optional[str] = None,
        thumb: typing.Optional[typing.Union['InputFile', str]] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendAudio", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_document(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        document: typing.Optional[typing.Union['InputFile', str]] = None,
        thumb: typing.Optional[typing.Union['InputFile', str]] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        disable_content_type_detection: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendDocument", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_video(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        video: typing.Optional[typing.Union['InputFile', str]] = None,
        duration: typing.Optional[int] = None,
        width: typing.Optional[int] = None,
        height: typing.Optional[int] = None,
        thumb: typing.Optional[typing.Union['InputFile', str]] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        supports_streaming: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendVideo", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_animation(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        animation: typing.Optional[typing.Union['InputFile', str]] = None,
        duration: typing.Optional[int] = None,
        width: typing.Optional[int] = None,
        height: typing.Optional[int] = None,
        thumb: typing.Optional[typing.Union['InputFile', str]] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendAnimation", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_voice(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        voice: typing.Optional[typing.Union['InputFile', str]] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        duration: typing.Optional[int] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendVoice", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_video_note(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        video_note: typing.Optional[typing.Union['InputFile', str]] = None,
        duration: typing.Optional[int] = None,
        length: typing.Optional[int] = None,
        thumb: typing.Optional[typing.Union['InputFile', str]] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendVideoNote", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_media_group(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        media: typing.Optional[typing.List[typing.Union['InputMediaAudio', 'InputMediaDocument', 'InputMediaPhoto', 'InputMediaVideo']]] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        **other
    ) -> Result[typing.List['Message'], APIError]:
        result = await self.api.request("sendMediaGroup", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=[Message(**self.get_response(e)) for e in u])
        return Result(False, error=result.error)

    async def send_location(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        latitude: typing.Optional[float] = None,
        longitude: typing.Optional[float] = None,
        horizontal_accuracy: typing.Optional[float] = None,
        live_period: typing.Optional[int] = None,
        heading: typing.Optional[int] = None,
        proximity_alert_radius: typing.Optional[int] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendLocation", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def edit_message_live_location(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        latitude: typing.Optional[float] = None,
        longitude: typing.Optional[float] = None,
        horizontal_accuracy: typing.Optional[float] = None,
        heading: typing.Optional[int] = None,
        proximity_alert_radius: typing.Optional[int] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("editMessageLiveLocation", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def stop_message_live_location(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("stopMessageLiveLocation", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_venue(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        latitude: typing.Optional[float] = None,
        longitude: typing.Optional[float] = None,
        title: typing.Optional[str] = None,
        address: typing.Optional[str] = None,
        foursquare_id: typing.Optional[str] = None,
        foursquare_type: typing.Optional[str] = None,
        google_place_id: typing.Optional[str] = None,
        google_place_type: typing.Optional[str] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendVenue", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_contact(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        phone_number: typing.Optional[str] = None,
        first_name: typing.Optional[str] = None,
        last_name: typing.Optional[str] = None,
        vcard: typing.Optional[str] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendContact", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_poll(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        question: typing.Optional[str] = None,
        options: typing.Optional[typing.List[str]] = None,
        is_anonymous: typing.Optional[bool] = None,
        type: typing.Optional[str] = None,
        allows_multiple_answers: typing.Optional[bool] = None,
        correct_option_id: typing.Optional[int] = None,
        explanation: typing.Optional[str] = None,
        explanation_parse_mode: typing.Optional[str] = None,
        explanation_entities: typing.Optional[typing.List['MessageEntity']] = None,
        open_period: typing.Optional[int] = None,
        close_date: typing.Optional[int] = None,
        is_closed: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendPoll", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_dice(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        emoji: typing.Optional[str] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendDice", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_chat_action(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        action: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("sendChatAction", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_user_profile_photos(
        self,
        user_id: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        **other
    ) -> Result['UserProfilePhotos', APIError]:
        result = await self.api.request("getUserProfilePhotos", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=UserProfilePhotos(**self.get_response(u)))
        return Result(False, error=result.error)

    async def get_file(
        self,
        file_id: typing.Optional[str] = None,
        **other
    ) -> Result['File', APIError]:
        result = await self.api.request("getFile", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=File(**self.get_response(u)))
        return Result(False, error=result.error)

    async def ban_chat_member(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        until_date: typing.Optional[int] = None,
        revoke_messages: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("banChatMember", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def unban_chat_member(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        only_if_banned: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("unbanChatMember", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def restrict_chat_member(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        permissions: typing.Optional['ChatPermissions'] = None,
        until_date: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("restrictChatMember", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def promote_chat_member(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        is_anonymous: typing.Optional[bool] = None,
        can_manage_chat: typing.Optional[bool] = None,
        can_post_messages: typing.Optional[bool] = None,
        can_edit_messages: typing.Optional[bool] = None,
        can_delete_messages: typing.Optional[bool] = None,
        can_manage_video_chats: typing.Optional[bool] = None,
        can_restrict_members: typing.Optional[bool] = None,
        can_promote_members: typing.Optional[bool] = None,
        can_change_info: typing.Optional[bool] = None,
        can_invite_users: typing.Optional[bool] = None,
        can_pin_messages: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("promoteChatMember", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_chat_administrator_custom_title(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        custom_title: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatAdministratorCustomTitle", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def ban_chat_sender_chat(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        sender_chat_id: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("banChatSenderChat", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def unban_chat_sender_chat(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        sender_chat_id: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("unbanChatSenderChat", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_chat_permissions(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        permissions: typing.Optional['ChatPermissions'] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatPermissions", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def export_chat_invite_link(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[str, APIError]:
        result = await self.api.request("exportChatInviteLink", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def create_chat_invite_link(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        name: typing.Optional[str] = None,
        expire_date: typing.Optional[int] = None,
        member_limit: typing.Optional[int] = None,
        creates_join_request: typing.Optional[bool] = None,
        **other
    ) -> Result['ChatInviteLink', APIError]:
        result = await self.api.request("createChatInviteLink", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=ChatInviteLink(**self.get_response(u)))
        return Result(False, error=result.error)

    async def edit_chat_invite_link(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        invite_link: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        expire_date: typing.Optional[int] = None,
        member_limit: typing.Optional[int] = None,
        creates_join_request: typing.Optional[bool] = None,
        **other
    ) -> Result['ChatInviteLink', APIError]:
        result = await self.api.request("editChatInviteLink", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=ChatInviteLink(**self.get_response(u)))
        return Result(False, error=result.error)

    async def revoke_chat_invite_link(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        invite_link: typing.Optional[str] = None,
        **other
    ) -> Result['ChatInviteLink', APIError]:
        result = await self.api.request("revokeChatInviteLink", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=ChatInviteLink(**self.get_response(u)))
        return Result(False, error=result.error)

    async def approve_chat_join_request(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("approveChatJoinRequest", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def decline_chat_join_request(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("declineChatJoinRequest", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_chat_photo(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        photo: typing.Optional['InputFile'] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatPhoto", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def delete_chat_photo(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("deleteChatPhoto", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_chat_title(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        title: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatTitle", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_chat_description(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        description: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatDescription", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def pin_chat_message(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        disable_notification: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("pinChatMessage", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def unpin_chat_message(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("unpinChatMessage", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def unpin_all_chat_messages(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("unpinAllChatMessages", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def leave_chat(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("leaveChat", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_chat(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result['Chat', APIError]:
        result = await self.api.request("getChat", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Chat(**self.get_response(u)))
        return Result(False, error=result.error)

    async def get_chat_administrators(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[typing.List['ChatMember'], APIError]:
        result = await self.api.request("getChatAdministrators", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=[ChatMember(**self.get_response(e)) for e in u])
        return Result(False, error=result.error)

    async def get_chat_member_count(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[int, APIError]:
        result = await self.api.request("getChatMemberCount", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_chat_member(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        user_id: typing.Optional[int] = None,
        **other
    ) -> Result['ChatMember', APIError]:
        result = await self.api.request("getChatMember", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=ChatMember(**self.get_response(u)))
        return Result(False, error=result.error)

    async def set_chat_sticker_set(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        sticker_set_name: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatStickerSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def delete_chat_sticker_set(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("deleteChatStickerSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def answer_callback_query(
        self,
        callback_query_id: typing.Optional[str] = None,
        text: typing.Optional[str] = None,
        show_alert: typing.Optional[bool] = None,
        url: typing.Optional[str] = None,
        cache_time: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("answerCallbackQuery", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_my_commands(
        self,
        commands: typing.Optional[typing.List['BotCommand']] = None,
        scope: typing.Optional['BotCommandScope'] = None,
        language_code: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setMyCommands", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def delete_my_commands(
        self,
        scope: typing.Optional['BotCommandScope'] = None,
        language_code: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("deleteMyCommands", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_my_commands(
        self,
        scope: typing.Optional['BotCommandScope'] = None,
        language_code: typing.Optional[str] = None,
        **other
    ) -> Result[typing.List['BotCommand'], APIError]:
        result = await self.api.request("getMyCommands", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=[BotCommand(**self.get_response(e)) for e in u])
        return Result(False, error=result.error)

    async def set_chat_menu_button(
        self,
        chat_id: typing.Optional[int] = None,
        menu_button: typing.Optional['MenuButton'] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setChatMenuButton", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_chat_menu_button(
        self,
        chat_id: typing.Optional[int] = None,
        **other
    ) -> Result['MenuButton', APIError]:
        result = await self.api.request("getChatMenuButton", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=MenuButton(**self.get_response(u)))
        return Result(False, error=result.error)

    async def set_my_default_administrator_rights(
        self,
        rights: typing.Optional['ChatAdministratorRights'] = None,
        for_channels: typing.Optional[bool] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setMyDefaultAdministratorRights", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def get_my_default_administrator_rights(
        self,
        for_channels: typing.Optional[bool] = None,
        **other
    ) -> Result['ChatAdministratorRights', APIError]:
        result = await self.api.request("getMyDefaultAdministratorRights", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=ChatAdministratorRights(**self.get_response(u)))
        return Result(False, error=result.error)

    async def edit_message_text(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        text: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        entities: typing.Optional[typing.List['MessageEntity']] = None,
        disable_web_page_preview: typing.Optional[bool] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("editMessageText", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def edit_message_caption(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        caption: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        caption_entities: typing.Optional[typing.List['MessageEntity']] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("editMessageCaption", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def edit_message_media(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        media: typing.Optional['InputMedia'] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("editMessageMedia", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def edit_message_reply_markup(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("editMessageReplyMarkup", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def stop_poll(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result['Poll', APIError]:
        result = await self.api.request("stopPoll", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Poll(**self.get_response(u)))
        return Result(False, error=result.error)

    async def delete_message(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        message_id: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("deleteMessage", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def send_sticker(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        sticker: typing.Optional[typing.Union['InputFile', str]] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[typing.Union['InlineKeyboardMarkup', 'ReplyKeyboardMarkup', 'ReplyKeyboardRemove', 'ForceReply']] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendSticker", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def get_sticker_set(
        self,
        name: typing.Optional[str] = None,
        **other
    ) -> Result['StickerSet', APIError]:
        result = await self.api.request("getStickerSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=StickerSet(**self.get_response(u)))
        return Result(False, error=result.error)

    async def upload_sticker_file(
        self,
        user_id: typing.Optional[int] = None,
        png_sticker: typing.Optional['InputFile'] = None,
        **other
    ) -> Result['File', APIError]:
        result = await self.api.request("uploadStickerFile", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=File(**self.get_response(u)))
        return Result(False, error=result.error)

    async def create_new_sticker_set(
        self,
        user_id: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        title: typing.Optional[str] = None,
        png_sticker: typing.Optional[typing.Union['InputFile', str]] = None,
        tgs_sticker: typing.Optional['InputFile'] = None,
        webm_sticker: typing.Optional['InputFile'] = None,
        emojis: typing.Optional[str] = None,
        contains_masks: typing.Optional[bool] = None,
        mask_position: typing.Optional['MaskPosition'] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("createNewStickerSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def add_sticker_to_set(
        self,
        user_id: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        png_sticker: typing.Optional[typing.Union['InputFile', str]] = None,
        tgs_sticker: typing.Optional['InputFile'] = None,
        webm_sticker: typing.Optional['InputFile'] = None,
        emojis: typing.Optional[str] = None,
        mask_position: typing.Optional['MaskPosition'] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("addStickerToSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_sticker_position_in_set(
        self,
        sticker: typing.Optional[str] = None,
        position: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setStickerPositionInSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def delete_sticker_from_set(
        self,
        sticker: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("deleteStickerFromSet", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_sticker_set_thumb(
        self,
        name: typing.Optional[str] = None,
        user_id: typing.Optional[int] = None,
        thumb: typing.Optional[typing.Union['InputFile', str]] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setStickerSetThumb", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def answer_inline_query(
        self,
        inline_query_id: typing.Optional[str] = None,
        results: typing.Optional[typing.List['InlineQueryResult']] = None,
        cache_time: typing.Optional[int] = None,
        is_personal: typing.Optional[bool] = None,
        next_offset: typing.Optional[str] = None,
        switch_pm_text: typing.Optional[str] = None,
        switch_pm_parameter: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("answerInlineQuery", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def answer_web_app_query(
        self,
        web_app_query_id: typing.Optional[str] = None,
        result: typing.Optional['InlineQueryResult'] = None,
        **other
    ) -> Result['SentWebAppMessage', APIError]:
        result = await self.api.request("answerWebAppQuery", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=SentWebAppMessage(**self.get_response(u)))
        return Result(False, error=result.error)

    async def send_invoice(
        self,
        chat_id: typing.Optional[typing.Union[int, str]] = None,
        title: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        payload: typing.Optional[str] = None,
        provider_token: typing.Optional[str] = None,
        currency: typing.Optional[str] = None,
        prices: typing.Optional[typing.List['LabeledPrice']] = None,
        max_tip_amount: typing.Optional[int] = None,
        suggested_tip_amounts: typing.Optional[typing.List[int]] = None,
        start_parameter: typing.Optional[str] = None,
        provider_data: typing.Optional[str] = None,
        photo_url: typing.Optional[str] = None,
        photo_size: typing.Optional[int] = None,
        photo_width: typing.Optional[int] = None,
        photo_height: typing.Optional[int] = None,
        need_name: typing.Optional[bool] = None,
        need_phone_number: typing.Optional[bool] = None,
        need_email: typing.Optional[bool] = None,
        need_shipping_address: typing.Optional[bool] = None,
        send_phone_number_to_provider: typing.Optional[bool] = None,
        send_email_to_provider: typing.Optional[bool] = None,
        is_flexible: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendInvoice", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def answer_shipping_query(
        self,
        shipping_query_id: typing.Optional[str] = None,
        ok: typing.Optional[bool] = None,
        shipping_options: typing.Optional[typing.List['ShippingOption']] = None,
        error_message: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("answerShippingQuery", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def answer_pre_checkout_query(
        self,
        pre_checkout_query_id: typing.Optional[str] = None,
        ok: typing.Optional[bool] = None,
        error_message: typing.Optional[str] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("answerPreCheckoutQuery", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def set_passport_data_errors(
        self,
        user_id: typing.Optional[int] = None,
        errors: typing.Optional[typing.List['PassportElementError']] = None,
        **other
    ) -> Result[bool, APIError]:
        result = await self.api.request("setPassportDataErrors", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=u)
        return Result(False, error=result.error)

    async def send_game(
        self,
        chat_id: typing.Optional[int] = None,
        game_short_name: typing.Optional[str] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional['InlineKeyboardMarkup'] = None,
        **other
    ) -> Result['Message', APIError]:
        result = await self.api.request("sendGame", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def set_game_score(
        self,
        user_id: typing.Optional[int] = None,
        score: typing.Optional[int] = None,
        force: typing.Optional[bool] = None,
        disable_edit_message: typing.Optional[bool] = None,
        chat_id: typing.Optional[int] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        **other
    ) -> Result[typing.Union['Message', bool], APIError]:
        result = await self.api.request("setGameScore", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            if isinstance(u, bool): return Result(True, value=u)
            return Result(True, value=Message(**self.get_response(u)))
        return Result(False, error=result.error)

    async def get_game_high_scores(
        self,
        user_id: typing.Optional[int] = None,
        chat_id: typing.Optional[int] = None,
        message_id: typing.Optional[int] = None,
        inline_message_id: typing.Optional[str] = None,
        **other
    ) -> Result[typing.List['GameHighScore'], APIError]:
        result = await self.api.request("getGameHighScores", self.get_params(locals()))
        if result.is_ok:
            u = result.unwrap()
            return Result(True, value=[GameHighScore(**self.get_response(e)) for e in u])
        return Result(False, error=result.error)


import asyncio

from .abc import ABCDispatch
from telegrinder.bot.rules import ABCRule
from .handler import ABCHandler, FuncHandler
from telegrinder.types import Update
from telegrinder.api.abc import ABCAPI
from .view import ABCView, MessageView, CallbackQueryView
import typing

T = typing.TypeVar("T")

DEFAULT_DATACLASS = Update


class Dispatch(ABCDispatch):
    def __init__(self):
        self.default_handlers: typing.List[ABCHandler] = []
        self.loop = asyncio.get_event_loop()
        self.message = MessageView()
        self.callback_query = CallbackQueryView()
        self.views = ["message", "callback_query"]

    def handle(
        self,
        *rules: ABCRule,
        is_blocking: bool = True,
        dataclass: typing.Any = DEFAULT_DATACLASS,
    ):
        def wrapper(func: typing.Callable):
            self.default_handlers.append(
                FuncHandler(func, list(rules), is_blocking, dataclass)
            )
            return func

        return wrapper

    def get_views(self) -> typing.Iterator[ABCView]:
        for view_name in self.views:
            view = getattr(self, view_name)
            assert view, f"View {view_name} is undefined in dispatch"
            yield view

    def load(self, external: "Dispatch"):
        for view_name in self.views:
            view = getattr(self, view_name)
            assert view, f"View {view_name} is undefined in dispatch"
            view_external = getattr(external, view_name)
            assert view_external, f"View {view_name} is undefined in external dispatch"
            view.load(view_external)

    async def feed(self, event: dict, api: ABCAPI) -> bool:

        for view in self.get_views():
            if await view.check(event):
                await view.process(event, api)
                return True

        found = False
        for handler in self.default_handlers:
            result = await handler.check(event)
            if result:
                found = True
                self.loop.create_task(handler.run(event))
                if handler.is_blocking:
                    return True
        return found



from .abc import ABCFormatter
import typing
from telegrinder.tools.parse_mode import ParseMode

FMT_CHARS = "\\`*_{}[]()#+-!"


def wrap_md(wrapper: str, s: str) -> "MarkdownFormatter":
    return MarkdownFormatter(wrapper + s + wrapper)


class MarkdownFormatter(ABCFormatter):
    PARSE_MODE = ParseMode.MARKDOWNV2

    def escape(self) -> "MarkdownFormatter":
        ns = ""
        for c in self:
            if c in FMT_CHARS:
                ns += "\\"
            ns += c
        ns = ns.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
        return MarkdownFormatter(ns)

    def escape_code(self) -> "MarkdownFormatter":
        return self.replace("\\", "\\\\").replace("`", "\\`")

    def escape_link(self) -> "MarkdownFormatter":
        return self.replace("`", "\\`").replace(")", "\\)")

    def bold(self) -> "MarkdownFormatter":
        return wrap_md("*", self)

    def italic(self) -> "MarkdownFormatter":
        return wrap_md("_", self)

    def underline(self) -> "MarkdownFormatter":
        return wrap_md("__", self)

    def strike(self) -> "MarkdownFormatter":
        return wrap_md("~", self)

    def link(self, href: str) -> "MarkdownFormatter":
        return MarkdownFormatter(
            f"[{self.escape()}]({MarkdownFormatter(href).escape_link()})"
        )

    def code_block(self) -> "MarkdownFormatter":
        return MarkdownFormatter(f"```\n{self.escape_code()}\n```")

    def code_block_with_lang(self, lang: str) -> "MarkdownFormatter":
        return MarkdownFormatter(
            f"```{MarkdownFormatter(lang).escape()}\n{self.escape_code()}\n```"
        )

    def code_inline(self) -> "MarkdownFormatter":
        return MarkdownFormatter(f"`{self.escape_code()}`")



from telegrinder.types import (
    Message,
    MessageEntity,
    InlineKeyboardMarkup,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
    ForceReply,
)
from telegrinder.types.methods import APIMethods
from telegrinder.api import API, APIError
from telegrinder.tools import Result
import typing


class MessageCute(Message):
    unprep_ctx_api: typing.Optional[typing.Any] = None

    @property
    def ctx_api(self) -> API:
        return getattr(self, "unprep_ctx_api")  # type: ignore

    async def answer(
        self,
        text: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        entities: typing.Optional[typing.List["MessageEntity"]] = None,
        disable_web_page_preview: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        reply_to_message_id: typing.Optional[int] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[
            typing.Union[
                "InlineKeyboardMarkup",
                "ReplyKeyboardMarkup",
                "ReplyKeyboardRemove",
                "ForceReply",
            ]
        ] = None,
        **other
    ) -> Result["Message", APIError]:
        params = APIMethods.get_params(locals())
        return await self.ctx_api.send_message(chat_id=self.chat.id, **params)

    async def reply(
        self,
        text: typing.Optional[str] = None,
        parse_mode: typing.Optional[str] = None,
        entities: typing.Optional[typing.List["MessageEntity"]] = None,
        disable_web_page_preview: typing.Optional[bool] = None,
        disable_notification: typing.Optional[bool] = None,
        protect_content: typing.Optional[bool] = None,
        allow_sending_without_reply: typing.Optional[bool] = None,
        reply_markup: typing.Optional[
            typing.Union[
                "InlineKeyboardMarkup",
                "ReplyKeyboardMarkup",
                "ReplyKeyboardRemove",
                "ForceReply",
            ]
        ] = None,
        **other
    ) -> Result["Message", APIError]:
        params = APIMethods.get_params(locals())
        return await self.ctx_api.send_message(
            chat_id=self.chat.id, reply_to_message_id=self.message_id, **params
        )



from abc import ABC, abstractmethod
import typing
from telegrinder.http_ import ABCClient
from telegrinder.tools import Result
from telegrinder.api.error import APIError

from envparse import env


class Token(str):
    @classmethod
    def from_env(cls, var_name: str = "BOT_TOKEN", is_read: bool = False) -> "Token":
        if not is_read:
            env.read_envfile()
        return cls(env.str(var_name))


class ABCAPI(ABC):
    http: ABCClient

    @abstractmethod
    async def request(
        self,
        method: str,
        data: typing.Optional[dict] = None,
    ) -> Result[typing.Union[list, dict, bool], APIError]:
        pass



from .abc import ABCScenario
from dataclasses import dataclass
from telegrinder.tools import InlineKeyboard, InlineButton
from telegrinder.types.objects import InlineKeyboardMarkup
from telegrinder.bot.cute_types import CallbackQueryCute
import typing
import random
import string

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch import Dispatch
    from telegrinder.api import API


@dataclass
class Choice:
    name: str
    is_picked: bool

    default_text: str
    picked_text: str
    code: str


def random_code(length: int) -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


class Checkbox(ABCScenario):
    INVALID_CODE: str = "Invalid code"
    CALLBACK_ANSWER: str = "Done"
    PARSE_MODE: str = "MarkdownV2"

    def __init__(
        self,
        chat_id: int,
        msg: str,
        multiple_choice: bool = True,
        ready_text: str = "Ready",
        max_in_row: int = 3,
    ):
        self.chat_id = chat_id
        self.msg = msg
        self.multiple_choice = multiple_choice
        self.choices: typing.List[Choice] = []
        self.ready = ready_text
        self.max_in_row = max_in_row
        self.random_code = random_code(16)

    def get_markup(self) -> InlineKeyboardMarkup:
        kb = InlineKeyboard(resize_keyboard=True)
        choices = self.choices.copy()
        while choices:
            while len(kb.keyboard[-1]) < self.max_in_row and choices:
                choice = choices.pop(0)
                kb.add(
                    InlineButton(
                        text=choice.default_text
                        if not choice.is_picked
                        else choice.picked_text,
                        callback_data=self.random_code + "/" + choice.code,
                    )
                )
            kb.row()
        kb.add(InlineButton(self.ready, callback_data=self.random_code + "/ready"))
        return kb.get_markup()

    def add_option(
        self, name: str, default_text: str, picked_text: str, is_picked: bool = False
    ) -> "Checkbox":
        self.choices.append(
            Choice(name, is_picked, default_text, picked_text, random_code(16))
        )
        return self

    async def handle(self, cb: CallbackQueryCute) -> bool:
        code = cb.data.replace(self.random_code + "/", "", 1)

        if code == "ready":
            return False

        for i, choice in enumerate(self.choices):
            if choice.code == code:
                # Toggle choice
                self.choices[i].is_picked = not self.choices[i].is_picked
                await cb.ctx_api.edit_message_text(
                    cb.message.chat.id,
                    cb.message.message_id,
                    text=self.msg,
                    parse_mode=self.PARSE_MODE,
                    reply_markup=self.get_markup(),
                )
                break

        return True

    async def wait(
        self, api: "API", dispatch: "Dispatch"
    ) -> typing.Tuple[typing.Dict[str, bool], int]:
        assert len(self.choices) > 1
        message = (
            await api.send_message(
                self.chat_id, self.msg, self.PARSE_MODE, reply_markup=self.get_markup()
            )
        ).unwrap()
        while True:
            q: CallbackQueryCute
            q, _ = await dispatch.callback_query.wait_for_answer(self.chat_id)
            should_continue = await self.handle(q)
            await q.answer(self.CALLBACK_ANSWER)
            if not should_continue:
                break
        return {
            choice.name: choice.is_picked for choice in self.choices
        }, message.message_id



import asyncio
import dataclasses
import typing
from telegrinder.bot.rules import ABCRule


DefaultWaiterHandler = typing.Callable[[typing.Any], typing.Coroutine]


@dataclasses.dataclass
class Waiter:
    rules: typing.Iterable[ABCRule]
    event: asyncio.Event
    default: typing.Optional[typing.Union[DefaultWaiterHandler, str]] = None


async def wait(waiter: Waiter) -> typing.Tuple[typing.Any, dict]:
    await waiter.event.wait()
    event, ctx = getattr(waiter.event, "e")
    return event, ctx



import logging
import os
import requests

URL = "https://ark0f.github.io/tg-bot-api/openapi.json"
TYPES = {
    "integer": "int",
    "string": "str",
    "long": "int",
    "bytes": "bytes",
    "boolean": "bool",
    "number": "float",
    "true": "bool",
    "false": "bool",
}
SPACES = "    "


def convert_type(d: dict) -> str:
    if "type" in d:
        t = d["type"]
        if t in TYPES:
            return TYPES[t]
        elif t == "array":
            nt = convert_type(d["items"])
            return "typing.List[" + nt + "]"
        else:
            if "." in t:
                t = t.split(".")[-1]
            return repr(t)
    elif "$ref" in d:
        n = d["$ref"].split("/")[-1]
        return repr(n)
    elif "anyOf" in d:
        return "typing.Union[" + ", ".join(convert_type(ut) for ut in d["anyOf"]) + "]"
    else:
        logging.error(f"cannot handle {d}")


def to_snakecase(s: str) -> str:
    ns = ""
    for i, symbol in enumerate(s):
        if i == 0:
            ns = ns + symbol.lower()
        else:
            ns = ns + (symbol if symbol.islower() else "_" + symbol.lower())
    return ns.replace("__", "_")


def get_lines_for_object(name: str, properties: dict):
    return [
        "\n\n",
        "class {}(BaseModel):\n".format(name),
        *(
            [SPACES + "pass\n"]
            if not properties
            else (
                SPACES
                + "{}: {}\n".format(
                    name if name not in ("json", "from") else name + "_",
                    "typing.Optional[" + convert_type(param) + "] = None",
                )
                for (name, param) in properties.items()
                if name != "flags"
            )
        ),
    ]


def parse_response(rt: str):
    if rt.startswith("'"):
        return "return Result(True, value=" + rt[1:-1] + "(**self.get_response(u)))"
    elif rt.startswith("typing"):
        if rt.startswith("typing.Union"):
            ts = rt[len("typing.Union") + 1 : -1].split(", ")
            ts_prim = [t for t in ts if not t.startswith("'")]
            s = ""
            for prim in ts_prim:
                s += f"if isinstance(u, {prim}): return Result(True, value=u)\n"
            comp = [t for t in ts if t not in ts_prim]
            if len(comp) > 1:
                print("cannot parse", rt)
                exit(0)
            s += (
                "return Result(True, value="
                + comp[0][1:-1]
                + "(**self.get_response(u)))"
            )
            return s
        elif rt.startswith("typing.List"):
            n = rt[len("typing.List") + 1 : -1]
            if not n.startswith("'"):
                print("no instruction to parse list of", n)
                exit(0)
            return f"return Result(True, value=[{n[1:-1]}(**self.get_response(e)) for e in u])"
    return "return Result(True, value=u)"


def generate(path: str, schema_url: str = URL) -> None:
    if not os.path.exists(path):
        os.makedirs(path)

    schema = requests.get(schema_url).json()

    paths = schema["paths"]
    objects = schema["components"]["schemas"]

    with open(path + "/__init__.py", "w") as file:
        file.writelines("from telegrinder.types.objects import *\n")

    with open(path + "/objects.py", "w") as file:
        file.writelines(
            ["import typing\n", "import inspect\n", "from telegrinder.model import *\n"]
        )

    for name, obj in objects.items():
        t, properties = obj.get("type", "object"), obj.get("properties", [])

        with open(path + "/objects.py", "a") as file:
            file.writelines(get_lines_for_object(name, properties))

    with open(path + "/objects.py", "a") as file:
        file.writelines(
            [
                "\n\n",
                "for v in locals().copy().values():\n",
                SPACES + "if inspect.isclass(v) and issubclass(v, BaseModel):\n",
                SPACES + SPACES + "v.update_forward_refs()",
                "\n\n",
                "__all__ = (\n",
            ]
            + [SPACES + repr(n) + ",\n" for n in objects]
            + [")\n"]
        )

    with open(path + "/methods.py", "w") as file:
        file.writelines(
            [
                "import typing\n",
                "from .objects import *\n",
                "from telegrinder.tools import Result\n",
                "from telegrinder.api.error import APIError\n\n",
                "if typing.TYPE_CHECKING:\n",
                SPACES + "from telegrinder.api.abc import ABCAPI\n\n",
                'X = typing.TypeVar("X")\n',
                "\n\n",
                "class APIMethods:\n",
                SPACES + 'def __init__(self, api: "ABCAPI"):\n',
                SPACES + SPACES + "self.api = api\n\n",
                SPACES + "@staticmethod\n",
                SPACES + "def get_params(loc: dict) -> dict:\n",
                SPACES
                + SPACES
                + 'n = {k: v for k, v in loc.items() if k not in ("self", "other") and v is not None}\n',
                SPACES + SPACES + "n.update(loc['other'])\n        return n\n\n",
                SPACES + "@staticmethod\n",
                SPACES + "def get_response(r: dict) -> dict:\n",
                SPACES + SPACES + "if 'json' in r: r['json_'] = r['json']\n",
                SPACES + SPACES + "return r",
            ]
        )

    for ps in paths:
        method = paths[ps]
        if "requestBody" not in method["post"]:
            props = {}
        else:
            props = list(method["post"]["requestBody"]["content"].values())[-1][
                "schema"
            ]["properties"]

        lines = []
        method_name = ps[1:]
        result = list(method["post"]["responses"]["200"]["content"].values())[-1][
            "schema"
        ]["properties"]["result"]
        response = convert_type(result)
        print(response)
        name = to_snakecase(method_name)
        lines.append(f"async def {name}(\n        self,\n")
        for n, prop in props.items():
            t = convert_type(prop)
            lines.append(SPACES + f"{n}: typing.Optional[{t}] = None,\n")
        lines.append(SPACES + "**other\n")
        lines.append(f") -> Result[{response}, APIError]:\n")
        lines.extend(
            [
                SPACES + li
                for li in (
                    "result = await self.api.request({}, self.get_params(locals()))\n".format(
                        '"' + method_name + '"'
                    ),
                    "if result.is_ok:\n",
                    SPACES + "u = result.unwrap()\n",
                    SPACES
                    + ("\n" + SPACES + SPACES + SPACES).join(
                        parse_response(response).split("\n")
                    )
                    + "\n",
                    "return Result(False, error=result.error)",
                )
            ]
        )
        with open(path + "/methods.py", "a") as file:
            file.writelines(["\n\n"] + [SPACES + li for li in lines])



from .abc import ABCAPI, APIError, Token
import typing
from telegrinder.tools import Result
from telegrinder.http_ import ABCClient, AiohttpClient
from telegrinder.types.methods import APIMethods
from telegrinder.model import convert
from telegrinder.modules import logger


def alias(d: typing.Any) -> typing.Any:
    if not isinstance(d, dict):
        if isinstance(d, list):
            return [alias(e) for e in d]
        return d
    for (k, v) in d.copy().items():
        if k in ("from",):
            d[k + "_"] = alias(d.pop(k))
        else:
            d[k] = alias(d[k])
    return d


class API(ABCAPI, APIMethods):
    API_URL = "https://api.telegram.org/"

    def __init__(self, token: Token, http: typing.Optional[ABCClient] = None):
        self.token = token
        self._request_url = self.API_URL + f"bot{self.token}/"
        self.http = http or AiohttpClient()
        super().__init__(self)

    async def request(
        self,
        method: str,
        data: typing.Optional[dict] = None,
    ) -> Result[typing.Union[dict, list, bool], APIError]:
        data = convert(data)
        logger.debug("Making API request {}: {}".format(method, data))
        response = await self.http.request_json(self._request_url + method, json=data)
        logger.debug("Got response: {}".format(response))
        if response.get("ok"):
            assert "result" in response
            return Result(True, value=alias(response["result"]))

        code, msg = response.get("error_code"), response.get("description")
        return Result(False, error=APIError(code, msg))



from .result import Result
from .keyboard import Keyboard, Button, InlineButton, InlineKeyboard, AnyMarkup
from .magic import VarUnset, resolve_arg_names, magic_bundle
from .kb_set import KeyboardSetBase, KeyboardSetYAML
from .parse_mode import ParseMode, get_mention_link
from .formatting import ABCFormatter, HTMLFormatter, MarkdownFormatter



from abc import ABC, abstractmethod
import typing


class ABCClient(ABC):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> dict:
        pass

    @abstractmethod
    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass

    async def __aenter__(self) -> "ABCClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()



from .abc import ABCDispatch
from .dispatch import ABCDispatch, Dispatch, ABCRule
from .handler import ABCHandler, FuncHandler
from .middleware import ABCMiddleware
from .view import ABCView, MessageView, CallbackQueryView



from .abc import ABCView
from telegrinder.bot.dispatch.handler import ABCHandler, FuncHandler
from telegrinder.bot.dispatch.waiter import Waiter, wait
from telegrinder.bot.dispatch.middleware import ABCMiddleware
from telegrinder.bot.rules import ABCRule
from telegrinder.bot.cute_types import MessageCute
from telegrinder.api.abc import ABCAPI
from telegrinder.bot.dispatch.waiter import DefaultWaiterHandler
import typing
import asyncio


class MessageView(ABCView):
    def __init__(self):
        self.handlers: typing.List[ABCHandler[MessageCute]] = []
        self.middlewares: typing.List[ABCMiddleware[MessageCute]] = []
        self.short_waiters: typing.Dict[int, Waiter] = {}
        self.loop = asyncio.get_event_loop()

    def __call__(self, *rules: ABCRule, is_blocking: bool = True):
        def wrapper(func: typing.Callable[..., typing.Coroutine]):
            self.handlers.append(
                FuncHandler(func, list(rules), is_blocking, dataclass=None)
            )
            return func
        return wrapper

    def load(self, external: "MessageView"):
        self.handlers.extend(external.handlers)
        self.middlewares.extend(external.middlewares)

    async def check(self, event: dict) -> bool:
        return "message" in event

    async def process(self, event: dict, api: ABCAPI):
        msg = MessageCute(**event["message"], unprep_ctx_api=api)

        if msg.chat.id in self.short_waiters:
            waiter = self.short_waiters[msg.chat.id]
            ctx = {}

            for rule in waiter.rules:
                chk_event = msg
                if rule.__event__ is None:
                    chk_event = event
                if not await rule.check(chk_event, ctx):
                    if not waiter.default:
                        return
                    elif isinstance(waiter.default, str):
                        await msg.answer(waiter.default)
                    else:
                        await waiter.default(msg)
                    return

            self.short_waiters.pop(msg.chat.id)
            setattr(waiter.event, "e", (msg, ctx))
            waiter.event.set()
            return

        ctx = {}

        for middleware in self.middlewares:
            if await middleware.pre(msg, ctx) is False:
                return False

        found = False
        responses = []
        for handler in self.handlers:
            result = await handler.check(event)
            if result:
                handler.ctx.update(ctx)
                found = True
                response = await handler.run(msg)
                responses.append(response)
                if handler.is_blocking:
                    break

        for middleware in self.middlewares:
            await middleware.post(msg, responses, ctx)

        return found

    async def wait_for_message(
        self,
        chat_id: int,
        *rules: ABCRule,
        default: typing.Optional[typing.Union[DefaultWaiterHandler, str]] = None
    ) -> typing.Tuple[MessageCute, dict]:
        event = asyncio.Event()
        waiter = Waiter(rules, event, default)
        self.short_waiters[chat_id] = waiter
        return await wait(waiter)



import pydantic
import typing


def convert(d: typing.Any) -> typing.Any:
    if isinstance(d, BaseModel):
        return d.get_dict()
    elif isinstance(d, dict):
        return {k: convert(v) for k, v in d.items() if v is not None}
    elif isinstance(d, list):
        li = [convert(el) for el in d]
        return li
    return d


class BaseModel(pydantic.BaseModel):
    def get_dict(self) -> dict:
        d = self.dict()
        return {k: convert(v) for k, v in d.items() if v is not None}


__all__ = ("convert", "BaseModel")



from telegrinder.types import CallbackQuery
from telegrinder.types.methods import APIMethods
from telegrinder.api import API, APIError
from telegrinder.tools import Result
import typing


class CallbackQueryCute(CallbackQuery):
    unprep_ctx_api: typing.Optional[typing.Any] = None

    @property
    def ctx_api(self) -> API:
        return getattr(self, "unprep_ctx_api")  # type: ignore

    async def answer(
        self,
        text: typing.Optional[str] = None,
        show_alert: typing.Optional[bool] = None,
        url: typing.Optional[str] = None,
        cache_time: typing.Optional[int] = None,
        **other
    ) -> Result[bool, APIError]:
        params = APIMethods.get_params(locals())
        return await self.ctx_api.answer_callback_query(self.id, **params)



import typing
import types

AnyMock = type("Any", (object,), {"__str__": lambda _: "Any"})


class VarUnset(BaseException):
    def __init__(self, name: str, t: typing.Any, handler: types.FunctionType):
        self.name = name
        self.t = t
        self.handler = handler

    def __repr__(self):
        return (
            f"Handler {self.handler.__name__} requires variable {self.name} of type {self.t} "
            "which was not set in the context"
        )

    def __str__(self):
        return self.__repr__()


def resolve_arg_names(func: types.FunctionType) -> typing.Tuple[str, ...]:
    return func.__code__.co_varnames[1 : func.__code__.co_argcount]


def magic_bundle(
    handler: types.FunctionType, kw: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    arg_names = resolve_arg_names(handler)
    hints = typing.get_type_hints(handler)
    kw_new = {}
    for name in arg_names:
        if name not in kw:
            raise VarUnset(name, hints.get(name, AnyMock).__name__, handler)
        kw_new[name] = kw[name]
    return kw_new



from abc import ABC, abstractmethod
from telegrinder.modules import json
from telegrinder.bot.cute_types import MessageCute
import typing
import collections
import vbml

T = typing.TypeVar("T")
patcher = vbml.Patcher()

Message = MessageCute
EventScheme = collections.namedtuple("EventScheme", ["name", "dataclass"])


class ABCRule(ABC, typing.Generic[T]):
    __event__: typing.Optional[EventScheme] = None

    @abstractmethod
    async def check(self, event: T, ctx: dict) -> bool:
        pass

    def __and__(self, other: "ABCRule"):
        return AndRule(self, other)

    def __or__(self, other: "ABCRule"):
        return OrRule(self, other)


class AndRule(ABCRule):
    def __init__(self, *rules: ABCRule):
        self.rules = rules

    async def check(self, event: dict, ctx: dict) -> bool:
        ctx_copy = ctx.copy()
        for rule in self.rules:
            e = event
            if rule.__event__:
                if rule.__event__.name not in event:
                    return False
                e = rule.__event__.dataclass(**event[rule.__event__.name])
            if not await rule.check(e, ctx_copy):
                return False
        ctx.clear()
        ctx.update(ctx_copy)
        return True


class OrRule(ABCRule):
    def __init__(self, *rules: ABCRule):
        self.rules = rules

    async def check(self, event: T, ctx: dict) -> bool:
        ctx_copy = ctx.copy()
        found = False

        for rule in self.rules:
            e = event
            if rule.__event__:
                if rule.__event__.name not in event:
                    continue
                e = rule.__event__.dataclass(**event[rule.__event__.name])
            if await rule.check(e, ctx_copy):
                found = True
                break

        ctx.clear()
        ctx.update(ctx_copy)
        return found


class ABCMessageRule(ABCRule, ABC):
    __event__ = EventScheme("message", Message)

    @abstractmethod
    async def check(self, message: Message, ctx: dict) -> bool:
        ...


class IsMessage(ABCRule):
    async def check(self, event: dict, ctx: dict) -> bool:
        return "message" in event


class Text(ABCMessageRule):
    def __init__(
        self, texts: typing.Union[str, typing.List[str]], ignore_case: bool = False
    ):
        if not isinstance(texts, list):
            texts = [texts]
        self.texts = texts
        self.ignore_case = ignore_case

    async def check(self, message: Message, ctx: dict) -> bool:
        if self.ignore_case:
            return message.text.lower() in list(map(str.lower, self.texts))
        return message.text in self.texts


class IsPrivate(ABCMessageRule):
    async def check(self, message: Message, ctx: dict) -> bool:
        return message.chat.id > 0


class IsChat(ABCRule):
    async def check(self, message: Message, ctx: dict) -> bool:
        return message.chat.id < 0


class Markup(ABCMessageRule):
    def __init__(self, patterns: typing.Union[str, typing.List[str]]):
        if not isinstance(patterns, list):
            patterns = [patterns]
        self.patterns = [
            vbml.Pattern(pattern) if isinstance(pattern, str) else pattern
            for pattern in patterns
        ]

    async def check(self, message: Message, ctx: dict) -> bool:
        for pattern in self.patterns:
            response = patcher.check(pattern, message.text)
            if response is False:
                continue
            ctx.update(response)
            return True
        return False


class FuncRule(ABCRule, typing.Generic[T]):
    def __init__(
        self,
        func: typing.Callable[[T, dict], bool],
        event_scheme: typing.Optional[
            typing.Union[
                EventScheme,
                typing.Tuple[str, typing.Type[T]]
            ]
        ] = None,
    ):
        self.func = func
        if isinstance(event_scheme, tuple):
            event_scheme = EventScheme(*event_scheme)
        self.event_scheme = event_scheme

    async def check(self, event: dict, ctx: dict) -> bool:
        if self.event_scheme:
            if self.event_scheme.name not in event:
                return False
            event = self.event_scheme.dataclass(**event[self.event_scheme.name])
        return self.func(event, ctx)


class CallbackDataEq(ABCRule):
    def __init__(self, value: str):
        self.value = value

    async def check(self, event: dict, ctx: dict) -> bool:
        return event["callback_query"].get("data") == self.value


class CallbackDataJsonEq(ABCRule):
    def __init__(self, d: dict):
        self.d = d

    async def check(self, event: dict, ctx: dict) -> bool:
        if "data" not in event["callback_query"]:
            return False
        try:
            return json.loads(event["callback_query"]["data"]) == self.d
        except:
            return False


__all__ = (
    ABCRule,
    ABCMessageRule,
    AndRule,
    OrRule,
    IsMessage,
    Text,
    Markup,
    IsPrivate,
    IsChat,
    FuncRule,
    CallbackDataEq,
    CallbackDataJsonEq,
)



from abc import ABC, abstractmethod
from telegrinder.tools.parse_mode import get_mention_link
import typing

T = typing.TypeVar("T")


class ABCFormatter(ABC, str):
    PARSE_MODE: str

    @abstractmethod
    def escape(self: T) -> T:
        ...

    @abstractmethod
    def bold(self: T) -> T:
        ...

    @abstractmethod
    def italic(self: T) -> T:
        ...

    @abstractmethod
    def underline(self: T) -> T:
        ...

    @abstractmethod
    def strike(self: T) -> T:
        ...

    @abstractmethod
    def link(self: T, href: str) -> T:
        ...

    def mention(self: T, user_id: int) -> T:
        return self.link(get_mention_link(user_id))

    @abstractmethod
    def code_block(self: T) -> T:
        ...

    @abstractmethod
    def code_block_with_lang(self: T, lang: str) -> T:
        ...

    @abstractmethod
    def code_inline(self: T) -> T:
        ...



import asyncio

from telegrinder.api import API
from telegrinder.bot.polling import ABCPolling, Polling
from telegrinder.bot.dispatch import ABCDispatch, Dispatch
import typing


class Telegrinder:
    def __init__(
        self,
        api: API,
        polling: typing.Optional[ABCPolling] = None,
        dispatch: typing.Optional[ABCDispatch] = None,
    ):
        self.api = api
        self.polling = polling or Polling(api)
        self.dispatch = dispatch or Dispatch()

    @property
    def on(self) -> Dispatch:
        return self.dispatch  # type: ignore

    async def run_polling(self, offset: int = 0) -> None:
        self.polling.offset = offset
        async for update in self.polling.listen():
            self.dispatch.loop.create_task(self.dispatch.feed(update, self.api))

    def run_forever(self, offset: int = 0) -> None:
        loop = asyncio.get_event_loop()
        loop.create_task(self.run_polling(offset))
        loop.run_forever()



from .polling import ABCPolling, Polling
from .dispatch import (
    ABCDispatch,
    Dispatch,
    ABCHandler,
    ABCView,
    FuncHandler,
    MessageView,
    CallbackQueryView,
    ABCMiddleware,
)
from .bot import Telegrinder
from .cute_types import MessageCute, CallbackQueryCute
from .rules import ABCRule, ABCMessageRule
from .scenario import ABCScenario, Checkbox



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['telegrinder',
 'telegrinder.api',
 'telegrinder.bot',
 'telegrinder.bot.cute_types',
 'telegrinder.bot.dispatch',
 'telegrinder.bot.dispatch.handler',
 'telegrinder.bot.dispatch.middleware',
 'telegrinder.bot.dispatch.view',
 'telegrinder.bot.polling',
 'telegrinder.bot.scenario',
 'telegrinder.http_',
 'telegrinder.tools',
 'telegrinder.tools.formatting',
 'telegrinder.tools.kb_set',
 'telegrinder.types']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'aiohttp>=3.8.1,<4.0.0',
 'choicelib>=0.1.5,<0.2.0',
 'envparse>=0.2.0,<0.3.0',
 'pydantic>=1.9.0,<2.0.0',
 'vbml>=1.1.post1,<2.0']

setup_kwargs = {
    'name': 'telegrinder',
    'version': '0.1.dev5',
    'description': 'Grinding telegram events – customizable tool to build telegram bots',
    'long_description': None,
    'author': 'timoniq',
    'author_email': 'tesseradecades@mail.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .http_ import ABCClient, AiohttpClient
from .api import ABCAPI, Token, API
from .bot import (
    ABCPolling,
    Polling,
    ABCDispatch,
    ABCRule,
    ABCMessageRule,
    Dispatch,
    Telegrinder,
    ABCView,
    ABCHandler,
    MessageView,
    CallbackQueryView,
    FuncHandler,
    MessageCute,
    CallbackQueryCute,
    ABCMiddleware,
    ABCScenario,
    Checkbox,
)
from .tools import (
    Result,
    Keyboard,
    Button,
    InlineButton,
    InlineKeyboard,
    VarUnset,
    magic_bundle,
    KeyboardSetBase,
    KeyboardSetYAML,
    AnyMarkup,
)

Message = MessageCute
CallbackQuery = CallbackQueryCute



import typing
import inspect
from telegrinder.model import *


class Error(BaseModel):
    ok: typing.Optional[bool] = None
    error_code: typing.Optional[int] = None
    description: typing.Optional[str] = None
    parameters: typing.Optional['ResponseParameters'] = None


class Update(BaseModel):
    update_id: typing.Optional[int] = None
    message: typing.Optional['Message'] = None
    edited_message: typing.Optional['Message'] = None
    channel_post: typing.Optional['Message'] = None
    edited_channel_post: typing.Optional['Message'] = None
    inline_query: typing.Optional['InlineQuery'] = None
    chosen_inline_result: typing.Optional['ChosenInlineResult'] = None
    callback_query: typing.Optional['CallbackQuery'] = None
    shipping_query: typing.Optional['ShippingQuery'] = None
    pre_checkout_query: typing.Optional['PreCheckoutQuery'] = None
    poll: typing.Optional['Poll'] = None
    poll_answer: typing.Optional['PollAnswer'] = None
    my_chat_member: typing.Optional['ChatMemberUpdated'] = None
    chat_member: typing.Optional['ChatMemberUpdated'] = None
    chat_join_request: typing.Optional['ChatJoinRequest'] = None


class WebhookInfo(BaseModel):
    url: typing.Optional[str] = None
    has_custom_certificate: typing.Optional[bool] = None
    pending_update_count: typing.Optional[int] = None
    ip_address: typing.Optional[str] = None
    last_error_date: typing.Optional[int] = None
    last_error_message: typing.Optional[str] = None
    last_synchronization_error_date: typing.Optional[int] = None
    max_connections: typing.Optional[int] = None
    allowed_updates: typing.Optional[typing.List[str]] = None


class User(BaseModel):
    id: typing.Optional[int] = None
    is_bot: typing.Optional[bool] = None
    first_name: typing.Optional[str] = None
    last_name: typing.Optional[str] = None
    username: typing.Optional[str] = None
    language_code: typing.Optional[str] = None
    can_join_groups: typing.Optional[bool] = None
    can_read_all_group_messages: typing.Optional[bool] = None
    supports_inline_queries: typing.Optional[bool] = None


class Chat(BaseModel):
    id: typing.Optional[int] = None
    type: typing.Optional[str] = None
    title: typing.Optional[str] = None
    username: typing.Optional[str] = None
    first_name: typing.Optional[str] = None
    last_name: typing.Optional[str] = None
    photo: typing.Optional['ChatPhoto'] = None
    bio: typing.Optional[str] = None
    has_private_forwards: typing.Optional[bool] = None
    description: typing.Optional[str] = None
    invite_link: typing.Optional[str] = None
    pinned_message: typing.Optional['Message'] = None
    permissions: typing.Optional['ChatPermissions'] = None
    slow_mode_delay: typing.Optional[int] = None
    message_auto_delete_time: typing.Optional[int] = None
    has_protected_content: typing.Optional[bool] = None
    sticker_set_name: typing.Optional[str] = None
    can_set_sticker_set: typing.Optional[bool] = None
    linked_chat_id: typing.Optional[int] = None
    location: typing.Optional['ChatLocation'] = None


class Message(BaseModel):
    message_id: typing.Optional[int] = None
    from_: typing.Optional['User'] = None
    sender_chat: typing.Optional['Chat'] = None
    date: typing.Optional[int] = None
    chat: typing.Optional['Chat'] = None
    forward_from: typing.Optional['User'] = None
    forward_from_chat: typing.Optional['Chat'] = None
    forward_from_message_id: typing.Optional[int] = None
    forward_signature: typing.Optional[str] = None
    forward_sender_name: typing.Optional[str] = None
    forward_date: typing.Optional[int] = None
    is_automatic_forward: typing.Optional[bool] = None
    reply_to_message: typing.Optional['Message'] = None
    via_bot: typing.Optional['User'] = None
    edit_date: typing.Optional[int] = None
    has_protected_content: typing.Optional[bool] = None
    media_group_id: typing.Optional[str] = None
    author_signature: typing.Optional[str] = None
    text: typing.Optional[str] = None
    entities: typing.Optional[typing.List['MessageEntity']] = None
    animation: typing.Optional['Animation'] = None
    audio: typing.Optional['Audio'] = None
    document: typing.Optional['Document'] = None
    photo: typing.Optional[typing.List['PhotoSize']] = None
    sticker: typing.Optional['Sticker'] = None
    video: typing.Optional['Video'] = None
    video_note: typing.Optional['VideoNote'] = None
    voice: typing.Optional['Voice'] = None
    caption: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    contact: typing.Optional['Contact'] = None
    dice: typing.Optional['Dice'] = None
    game: typing.Optional['Game'] = None
    poll: typing.Optional['Poll'] = None
    venue: typing.Optional['Venue'] = None
    location: typing.Optional['Location'] = None
    new_chat_members: typing.Optional[typing.List['User']] = None
    left_chat_member: typing.Optional['User'] = None
    new_chat_title: typing.Optional[str] = None
    new_chat_photo: typing.Optional[typing.List['PhotoSize']] = None
    delete_chat_photo: typing.Optional[bool] = None
    group_chat_created: typing.Optional[bool] = None
    supergroup_chat_created: typing.Optional[bool] = None
    channel_chat_created: typing.Optional[bool] = None
    message_auto_delete_timer_changed: typing.Optional['MessageAutoDeleteTimerChanged'] = None
    migrate_to_chat_id: typing.Optional[int] = None
    migrate_from_chat_id: typing.Optional[int] = None
    pinned_message: typing.Optional['Message'] = None
    invoice: typing.Optional['Invoice'] = None
    successful_payment: typing.Optional['SuccessfulPayment'] = None
    connected_website: typing.Optional[str] = None
    passport_data: typing.Optional['PassportData'] = None
    proximity_alert_triggered: typing.Optional['ProximityAlertTriggered'] = None
    video_chat_scheduled: typing.Optional['VideoChatScheduled'] = None
    video_chat_started: typing.Optional['VideoChatStarted'] = None
    video_chat_ended: typing.Optional['VideoChatEnded'] = None
    video_chat_participants_invited: typing.Optional['VideoChatParticipantsInvited'] = None
    web_app_data: typing.Optional['WebAppData'] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None


class MessageId(BaseModel):
    message_id: typing.Optional[int] = None


class MessageEntity(BaseModel):
    type: typing.Optional[str] = None
    offset: typing.Optional[int] = None
    length: typing.Optional[int] = None
    url: typing.Optional[str] = None
    user: typing.Optional['User'] = None
    language: typing.Optional[str] = None


class PhotoSize(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    file_size: typing.Optional[int] = None


class Animation(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    duration: typing.Optional[int] = None
    thumb: typing.Optional['PhotoSize'] = None
    file_name: typing.Optional[str] = None
    mime_type: typing.Optional[str] = None
    file_size: typing.Optional[int] = None


class Audio(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    duration: typing.Optional[int] = None
    performer: typing.Optional[str] = None
    title: typing.Optional[str] = None
    file_name: typing.Optional[str] = None
    mime_type: typing.Optional[str] = None
    file_size: typing.Optional[int] = None
    thumb: typing.Optional['PhotoSize'] = None


class Document(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    thumb: typing.Optional['PhotoSize'] = None
    file_name: typing.Optional[str] = None
    mime_type: typing.Optional[str] = None
    file_size: typing.Optional[int] = None


class Video(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    duration: typing.Optional[int] = None
    thumb: typing.Optional['PhotoSize'] = None
    file_name: typing.Optional[str] = None
    mime_type: typing.Optional[str] = None
    file_size: typing.Optional[int] = None


class VideoNote(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    length: typing.Optional[int] = None
    duration: typing.Optional[int] = None
    thumb: typing.Optional['PhotoSize'] = None
    file_size: typing.Optional[int] = None


class Voice(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    duration: typing.Optional[int] = None
    mime_type: typing.Optional[str] = None
    file_size: typing.Optional[int] = None


class Contact(BaseModel):
    phone_number: typing.Optional[str] = None
    first_name: typing.Optional[str] = None
    last_name: typing.Optional[str] = None
    user_id: typing.Optional[int] = None
    vcard: typing.Optional[str] = None


class Dice(BaseModel):
    emoji: typing.Optional[str] = None
    value: typing.Optional[int] = None


class PollOption(BaseModel):
    text: typing.Optional[str] = None
    voter_count: typing.Optional[int] = None


class PollAnswer(BaseModel):
    poll_id: typing.Optional[str] = None
    user: typing.Optional['User'] = None
    option_ids: typing.Optional[typing.List[int]] = None


class Poll(BaseModel):
    id: typing.Optional[str] = None
    question: typing.Optional[str] = None
    options: typing.Optional[typing.List['PollOption']] = None
    total_voter_count: typing.Optional[int] = None
    is_closed: typing.Optional[bool] = None
    is_anonymous: typing.Optional[bool] = None
    type: typing.Optional[str] = None
    allows_multiple_answers: typing.Optional[bool] = None
    correct_option_id: typing.Optional[int] = None
    explanation: typing.Optional[str] = None
    explanation_entities: typing.Optional[typing.List['MessageEntity']] = None
    open_period: typing.Optional[int] = None
    close_date: typing.Optional[int] = None


class Location(BaseModel):
    longitude: typing.Optional[float] = None
    latitude: typing.Optional[float] = None
    horizontal_accuracy: typing.Optional[float] = None
    live_period: typing.Optional[int] = None
    heading: typing.Optional[int] = None
    proximity_alert_radius: typing.Optional[int] = None


class Venue(BaseModel):
    location: typing.Optional['Location'] = None
    title: typing.Optional[str] = None
    address: typing.Optional[str] = None
    foursquare_id: typing.Optional[str] = None
    foursquare_type: typing.Optional[str] = None
    google_place_id: typing.Optional[str] = None
    google_place_type: typing.Optional[str] = None


class WebAppData(BaseModel):
    data: typing.Optional[str] = None
    button_text: typing.Optional[str] = None


class ProximityAlertTriggered(BaseModel):
    traveler: typing.Optional['User'] = None
    watcher: typing.Optional['User'] = None
    distance: typing.Optional[int] = None


class MessageAutoDeleteTimerChanged(BaseModel):
    message_auto_delete_time: typing.Optional[int] = None


class VideoChatScheduled(BaseModel):
    start_date: typing.Optional[int] = None


class VideoChatStarted(BaseModel):
    pass


class VideoChatEnded(BaseModel):
    duration: typing.Optional[int] = None


class VideoChatParticipantsInvited(BaseModel):
    users: typing.Optional[typing.List['User']] = None


class UserProfilePhotos(BaseModel):
    total_count: typing.Optional[int] = None
    photos: typing.Optional[typing.List[typing.List['PhotoSize']]] = None


class File(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    file_size: typing.Optional[int] = None
    file_path: typing.Optional[str] = None


class WebAppInfo(BaseModel):
    url: typing.Optional[str] = None


class ReplyKeyboardMarkup(BaseModel):
    keyboard: typing.Optional[typing.List[typing.List['KeyboardButton']]] = None
    resize_keyboard: typing.Optional[bool] = None
    one_time_keyboard: typing.Optional[bool] = None
    input_field_placeholder: typing.Optional[str] = None
    selective: typing.Optional[bool] = None


class KeyboardButton(BaseModel):
    text: typing.Optional[str] = None
    request_contact: typing.Optional[bool] = None
    request_location: typing.Optional[bool] = None
    request_poll: typing.Optional['KeyboardButtonPollType'] = None
    web_app: typing.Optional['WebAppInfo'] = None


class KeyboardButtonPollType(BaseModel):
    type: typing.Optional[str] = None


class ReplyKeyboardRemove(BaseModel):
    remove_keyboard: typing.Optional[bool] = None
    selective: typing.Optional[bool] = None


class InlineKeyboardMarkup(BaseModel):
    inline_keyboard: typing.Optional[typing.List[typing.List['InlineKeyboardButton']]] = None


class InlineKeyboardButton(BaseModel):
    text: typing.Optional[str] = None
    url: typing.Optional[str] = None
    callback_data: typing.Optional[str] = None
    web_app: typing.Optional['WebAppInfo'] = None
    login_url: typing.Optional['LoginUrl'] = None
    switch_inline_query: typing.Optional[str] = None
    switch_inline_query_current_chat: typing.Optional[str] = None
    callback_game: typing.Optional['CallbackGame'] = None
    pay: typing.Optional[bool] = None


class LoginUrl(BaseModel):
    url: typing.Optional[str] = None
    forward_text: typing.Optional[str] = None
    bot_username: typing.Optional[str] = None
    request_write_access: typing.Optional[bool] = None


class CallbackQuery(BaseModel):
    id: typing.Optional[str] = None
    from_: typing.Optional['User'] = None
    message: typing.Optional['Message'] = None
    inline_message_id: typing.Optional[str] = None
    chat_instance: typing.Optional[str] = None
    data: typing.Optional[str] = None
    game_short_name: typing.Optional[str] = None


class ForceReply(BaseModel):
    force_reply: typing.Optional[bool] = None
    input_field_placeholder: typing.Optional[str] = None
    selective: typing.Optional[bool] = None


class ChatPhoto(BaseModel):
    small_file_id: typing.Optional[str] = None
    small_file_unique_id: typing.Optional[str] = None
    big_file_id: typing.Optional[str] = None
    big_file_unique_id: typing.Optional[str] = None


class ChatInviteLink(BaseModel):
    invite_link: typing.Optional[str] = None
    creator: typing.Optional['User'] = None
    creates_join_request: typing.Optional[bool] = None
    is_primary: typing.Optional[bool] = None
    is_revoked: typing.Optional[bool] = None
    name: typing.Optional[str] = None
    expire_date: typing.Optional[int] = None
    member_limit: typing.Optional[int] = None
    pending_join_request_count: typing.Optional[int] = None


class ChatAdministratorRights(BaseModel):
    is_anonymous: typing.Optional[bool] = None
    can_manage_chat: typing.Optional[bool] = None
    can_delete_messages: typing.Optional[bool] = None
    can_manage_video_chats: typing.Optional[bool] = None
    can_restrict_members: typing.Optional[bool] = None
    can_promote_members: typing.Optional[bool] = None
    can_change_info: typing.Optional[bool] = None
    can_invite_users: typing.Optional[bool] = None
    can_post_messages: typing.Optional[bool] = None
    can_edit_messages: typing.Optional[bool] = None
    can_pin_messages: typing.Optional[bool] = None


class ChatMember(BaseModel):
    pass


class ChatMemberOwner(BaseModel):
    status: typing.Optional[str] = None
    user: typing.Optional['User'] = None
    is_anonymous: typing.Optional[bool] = None
    custom_title: typing.Optional[str] = None


class ChatMemberAdministrator(BaseModel):
    status: typing.Optional[str] = None
    user: typing.Optional['User'] = None
    can_be_edited: typing.Optional[bool] = None
    is_anonymous: typing.Optional[bool] = None
    can_manage_chat: typing.Optional[bool] = None
    can_delete_messages: typing.Optional[bool] = None
    can_manage_video_chats: typing.Optional[bool] = None
    can_restrict_members: typing.Optional[bool] = None
    can_promote_members: typing.Optional[bool] = None
    can_change_info: typing.Optional[bool] = None
    can_invite_users: typing.Optional[bool] = None
    can_post_messages: typing.Optional[bool] = None
    can_edit_messages: typing.Optional[bool] = None
    can_pin_messages: typing.Optional[bool] = None
    custom_title: typing.Optional[str] = None


class ChatMemberMember(BaseModel):
    status: typing.Optional[str] = None
    user: typing.Optional['User'] = None


class ChatMemberRestricted(BaseModel):
    status: typing.Optional[str] = None
    user: typing.Optional['User'] = None
    is_member: typing.Optional[bool] = None
    can_change_info: typing.Optional[bool] = None
    can_invite_users: typing.Optional[bool] = None
    can_pin_messages: typing.Optional[bool] = None
    can_send_messages: typing.Optional[bool] = None
    can_send_media_messages: typing.Optional[bool] = None
    can_send_polls: typing.Optional[bool] = None
    can_send_other_messages: typing.Optional[bool] = None
    can_add_web_page_previews: typing.Optional[bool] = None
    until_date: typing.Optional[int] = None


class ChatMemberLeft(BaseModel):
    status: typing.Optional[str] = None
    user: typing.Optional['User'] = None


class ChatMemberBanned(BaseModel):
    status: typing.Optional[str] = None
    user: typing.Optional['User'] = None
    until_date: typing.Optional[int] = None


class ChatMemberUpdated(BaseModel):
    chat: typing.Optional['Chat'] = None
    from_: typing.Optional['User'] = None
    date: typing.Optional[int] = None
    old_chat_member: typing.Optional['ChatMember'] = None
    new_chat_member: typing.Optional['ChatMember'] = None
    invite_link: typing.Optional['ChatInviteLink'] = None


class ChatJoinRequest(BaseModel):
    chat: typing.Optional['Chat'] = None
    from_: typing.Optional['User'] = None
    date: typing.Optional[int] = None
    bio: typing.Optional[str] = None
    invite_link: typing.Optional['ChatInviteLink'] = None


class ChatPermissions(BaseModel):
    can_send_messages: typing.Optional[bool] = None
    can_send_media_messages: typing.Optional[bool] = None
    can_send_polls: typing.Optional[bool] = None
    can_send_other_messages: typing.Optional[bool] = None
    can_add_web_page_previews: typing.Optional[bool] = None
    can_change_info: typing.Optional[bool] = None
    can_invite_users: typing.Optional[bool] = None
    can_pin_messages: typing.Optional[bool] = None


class ChatLocation(BaseModel):
    location: typing.Optional['Location'] = None
    address: typing.Optional[str] = None


class BotCommand(BaseModel):
    command: typing.Optional[str] = None
    description: typing.Optional[str] = None


class BotCommandScope(BaseModel):
    pass


class BotCommandScopeDefault(BaseModel):
    type: typing.Optional[str] = None


class BotCommandScopeAllPrivateChats(BaseModel):
    type: typing.Optional[str] = None


class BotCommandScopeAllGroupChats(BaseModel):
    type: typing.Optional[str] = None


class BotCommandScopeAllChatAdministrators(BaseModel):
    type: typing.Optional[str] = None


class BotCommandScopeChat(BaseModel):
    type: typing.Optional[str] = None
    chat_id: typing.Optional[typing.Union[int, str]] = None


class BotCommandScopeChatAdministrators(BaseModel):
    type: typing.Optional[str] = None
    chat_id: typing.Optional[typing.Union[int, str]] = None


class BotCommandScopeChatMember(BaseModel):
    type: typing.Optional[str] = None
    chat_id: typing.Optional[typing.Union[int, str]] = None
    user_id: typing.Optional[int] = None


class MenuButton(BaseModel):
    pass


class MenuButtonCommands(BaseModel):
    type: typing.Optional[str] = None


class MenuButtonWebApp(BaseModel):
    type: typing.Optional[str] = None
    text: typing.Optional[str] = None
    web_app: typing.Optional['WebAppInfo'] = None


class MenuButtonDefault(BaseModel):
    type: typing.Optional[str] = None


class ResponseParameters(BaseModel):
    migrate_to_chat_id: typing.Optional[int] = None
    retry_after: typing.Optional[int] = None


class InputMedia(BaseModel):
    pass


class InputMediaPhoto(BaseModel):
    type: typing.Optional[str] = None
    media: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None


class InputMediaVideo(BaseModel):
    type: typing.Optional[str] = None
    media: typing.Optional[str] = None
    thumb: typing.Optional[typing.Union['InputFile', str]] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    duration: typing.Optional[int] = None
    supports_streaming: typing.Optional[bool] = None


class InputMediaAnimation(BaseModel):
    type: typing.Optional[str] = None
    media: typing.Optional[str] = None
    thumb: typing.Optional[typing.Union['InputFile', str]] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    duration: typing.Optional[int] = None


class InputMediaAudio(BaseModel):
    type: typing.Optional[str] = None
    media: typing.Optional[str] = None
    thumb: typing.Optional[typing.Union['InputFile', str]] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    duration: typing.Optional[int] = None
    performer: typing.Optional[str] = None
    title: typing.Optional[str] = None


class InputMediaDocument(BaseModel):
    type: typing.Optional[str] = None
    media: typing.Optional[str] = None
    thumb: typing.Optional[typing.Union['InputFile', str]] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    disable_content_type_detection: typing.Optional[bool] = None


class InputFile(BaseModel):
    pass


class Sticker(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    width: typing.Optional[int] = None
    height: typing.Optional[int] = None
    is_animated: typing.Optional[bool] = None
    is_video: typing.Optional[bool] = None
    thumb: typing.Optional['PhotoSize'] = None
    emoji: typing.Optional[str] = None
    set_name: typing.Optional[str] = None
    mask_position: typing.Optional['MaskPosition'] = None
    file_size: typing.Optional[int] = None


class StickerSet(BaseModel):
    name: typing.Optional[str] = None
    title: typing.Optional[str] = None
    is_animated: typing.Optional[bool] = None
    is_video: typing.Optional[bool] = None
    contains_masks: typing.Optional[bool] = None
    stickers: typing.Optional[typing.List['Sticker']] = None
    thumb: typing.Optional['PhotoSize'] = None


class MaskPosition(BaseModel):
    point: typing.Optional[str] = None
    x_shift: typing.Optional[float] = None
    y_shift: typing.Optional[float] = None
    scale: typing.Optional[float] = None


class InlineQuery(BaseModel):
    id: typing.Optional[str] = None
    from_: typing.Optional['User'] = None
    query: typing.Optional[str] = None
    offset: typing.Optional[str] = None
    chat_type: typing.Optional[str] = None
    location: typing.Optional['Location'] = None


class InlineQueryResult(BaseModel):
    pass


class InlineQueryResultArticle(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    input_message_content: typing.Optional['InputMessageContent'] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    url: typing.Optional[str] = None
    hide_url: typing.Optional[bool] = None
    description: typing.Optional[str] = None
    thumb_url: typing.Optional[str] = None
    thumb_width: typing.Optional[int] = None
    thumb_height: typing.Optional[int] = None


class InlineQueryResultPhoto(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    photo_url: typing.Optional[str] = None
    thumb_url: typing.Optional[str] = None
    photo_width: typing.Optional[int] = None
    photo_height: typing.Optional[int] = None
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultGif(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    gif_url: typing.Optional[str] = None
    gif_width: typing.Optional[int] = None
    gif_height: typing.Optional[int] = None
    gif_duration: typing.Optional[int] = None
    thumb_url: typing.Optional[str] = None
    thumb_mime_type: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultMpeg4Gif(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    mpeg4_url: typing.Optional[str] = None
    mpeg4_width: typing.Optional[int] = None
    mpeg4_height: typing.Optional[int] = None
    mpeg4_duration: typing.Optional[int] = None
    thumb_url: typing.Optional[str] = None
    thumb_mime_type: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultVideo(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    video_url: typing.Optional[str] = None
    mime_type: typing.Optional[str] = None
    thumb_url: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    video_width: typing.Optional[int] = None
    video_height: typing.Optional[int] = None
    video_duration: typing.Optional[int] = None
    description: typing.Optional[str] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultAudio(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    audio_url: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    performer: typing.Optional[str] = None
    audio_duration: typing.Optional[int] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultVoice(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    voice_url: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    voice_duration: typing.Optional[int] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultDocument(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    document_url: typing.Optional[str] = None
    mime_type: typing.Optional[str] = None
    description: typing.Optional[str] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None
    thumb_url: typing.Optional[str] = None
    thumb_width: typing.Optional[int] = None
    thumb_height: typing.Optional[int] = None


class InlineQueryResultLocation(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    latitude: typing.Optional[float] = None
    longitude: typing.Optional[float] = None
    title: typing.Optional[str] = None
    horizontal_accuracy: typing.Optional[float] = None
    live_period: typing.Optional[int] = None
    heading: typing.Optional[int] = None
    proximity_alert_radius: typing.Optional[int] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None
    thumb_url: typing.Optional[str] = None
    thumb_width: typing.Optional[int] = None
    thumb_height: typing.Optional[int] = None


class InlineQueryResultVenue(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    latitude: typing.Optional[float] = None
    longitude: typing.Optional[float] = None
    title: typing.Optional[str] = None
    address: typing.Optional[str] = None
    foursquare_id: typing.Optional[str] = None
    foursquare_type: typing.Optional[str] = None
    google_place_id: typing.Optional[str] = None
    google_place_type: typing.Optional[str] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None
    thumb_url: typing.Optional[str] = None
    thumb_width: typing.Optional[int] = None
    thumb_height: typing.Optional[int] = None


class InlineQueryResultContact(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    phone_number: typing.Optional[str] = None
    first_name: typing.Optional[str] = None
    last_name: typing.Optional[str] = None
    vcard: typing.Optional[str] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None
    thumb_url: typing.Optional[str] = None
    thumb_width: typing.Optional[int] = None
    thumb_height: typing.Optional[int] = None


class InlineQueryResultGame(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    game_short_name: typing.Optional[str] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None


class InlineQueryResultCachedPhoto(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    photo_file_id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedGif(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    gif_file_id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedMpeg4Gif(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    mpeg4_file_id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedSticker(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    sticker_file_id: typing.Optional[str] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedDocument(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    document_file_id: typing.Optional[str] = None
    description: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedVideo(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    video_file_id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedVoice(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    voice_file_id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InlineQueryResultCachedAudio(BaseModel):
    type: typing.Optional[str] = None
    id: typing.Optional[str] = None
    audio_file_id: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    caption_entities: typing.Optional[typing.List['MessageEntity']] = None
    reply_markup: typing.Optional['InlineKeyboardMarkup'] = None
    input_message_content: typing.Optional['InputMessageContent'] = None


class InputMessageContent(BaseModel):
    pass


class InputTextMessageContent(BaseModel):
    message_text: typing.Optional[str] = None
    parse_mode: typing.Optional[str] = None
    entities: typing.Optional[typing.List['MessageEntity']] = None
    disable_web_page_preview: typing.Optional[bool] = None


class InputLocationMessageContent(BaseModel):
    latitude: typing.Optional[float] = None
    longitude: typing.Optional[float] = None
    horizontal_accuracy: typing.Optional[float] = None
    live_period: typing.Optional[int] = None
    heading: typing.Optional[int] = None
    proximity_alert_radius: typing.Optional[int] = None


class InputVenueMessageContent(BaseModel):
    latitude: typing.Optional[float] = None
    longitude: typing.Optional[float] = None
    title: typing.Optional[str] = None
    address: typing.Optional[str] = None
    foursquare_id: typing.Optional[str] = None
    foursquare_type: typing.Optional[str] = None
    google_place_id: typing.Optional[str] = None
    google_place_type: typing.Optional[str] = None


class InputContactMessageContent(BaseModel):
    phone_number: typing.Optional[str] = None
    first_name: typing.Optional[str] = None
    last_name: typing.Optional[str] = None
    vcard: typing.Optional[str] = None


class InputInvoiceMessageContent(BaseModel):
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None
    payload: typing.Optional[str] = None
    provider_token: typing.Optional[str] = None
    currency: typing.Optional[str] = None
    prices: typing.Optional[typing.List['LabeledPrice']] = None
    max_tip_amount: typing.Optional[int] = None
    suggested_tip_amounts: typing.Optional[typing.List[int]] = None
    provider_data: typing.Optional[str] = None
    photo_url: typing.Optional[str] = None
    photo_size: typing.Optional[int] = None
    photo_width: typing.Optional[int] = None
    photo_height: typing.Optional[int] = None
    need_name: typing.Optional[bool] = None
    need_phone_number: typing.Optional[bool] = None
    need_email: typing.Optional[bool] = None
    need_shipping_address: typing.Optional[bool] = None
    send_phone_number_to_provider: typing.Optional[bool] = None
    send_email_to_provider: typing.Optional[bool] = None
    is_flexible: typing.Optional[bool] = None


class ChosenInlineResult(BaseModel):
    result_id: typing.Optional[str] = None
    from_: typing.Optional['User'] = None
    location: typing.Optional['Location'] = None
    inline_message_id: typing.Optional[str] = None
    query: typing.Optional[str] = None


class SentWebAppMessage(BaseModel):
    inline_message_id: typing.Optional[str] = None


class LabeledPrice(BaseModel):
    label: typing.Optional[str] = None
    amount: typing.Optional[int] = None


class Invoice(BaseModel):
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None
    start_parameter: typing.Optional[str] = None
    currency: typing.Optional[str] = None
    total_amount: typing.Optional[int] = None


class ShippingAddress(BaseModel):
    country_code: typing.Optional[str] = None
    state: typing.Optional[str] = None
    city: typing.Optional[str] = None
    street_line1: typing.Optional[str] = None
    street_line2: typing.Optional[str] = None
    post_code: typing.Optional[str] = None


class OrderInfo(BaseModel):
    name: typing.Optional[str] = None
    phone_number: typing.Optional[str] = None
    email: typing.Optional[str] = None
    shipping_address: typing.Optional['ShippingAddress'] = None


class ShippingOption(BaseModel):
    id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    prices: typing.Optional[typing.List['LabeledPrice']] = None


class SuccessfulPayment(BaseModel):
    currency: typing.Optional[str] = None
    total_amount: typing.Optional[int] = None
    invoice_payload: typing.Optional[str] = None
    shipping_option_id: typing.Optional[str] = None
    order_info: typing.Optional['OrderInfo'] = None
    telegram_payment_charge_id: typing.Optional[str] = None
    provider_payment_charge_id: typing.Optional[str] = None


class ShippingQuery(BaseModel):
    id: typing.Optional[str] = None
    from_: typing.Optional['User'] = None
    invoice_payload: typing.Optional[str] = None
    shipping_address: typing.Optional['ShippingAddress'] = None


class PreCheckoutQuery(BaseModel):
    id: typing.Optional[str] = None
    from_: typing.Optional['User'] = None
    currency: typing.Optional[str] = None
    total_amount: typing.Optional[int] = None
    invoice_payload: typing.Optional[str] = None
    shipping_option_id: typing.Optional[str] = None
    order_info: typing.Optional['OrderInfo'] = None


class PassportData(BaseModel):
    data: typing.Optional[typing.List['EncryptedPassportElement']] = None
    credentials: typing.Optional['EncryptedCredentials'] = None


class PassportFile(BaseModel):
    file_id: typing.Optional[str] = None
    file_unique_id: typing.Optional[str] = None
    file_size: typing.Optional[int] = None
    file_date: typing.Optional[int] = None


class EncryptedPassportElement(BaseModel):
    type: typing.Optional[str] = None
    data: typing.Optional[str] = None
    phone_number: typing.Optional[str] = None
    email: typing.Optional[str] = None
    files: typing.Optional[typing.List['PassportFile']] = None
    front_side: typing.Optional['PassportFile'] = None
    reverse_side: typing.Optional['PassportFile'] = None
    selfie: typing.Optional['PassportFile'] = None
    translation: typing.Optional[typing.List['PassportFile']] = None
    hash: typing.Optional[str] = None


class EncryptedCredentials(BaseModel):
    data: typing.Optional[str] = None
    hash: typing.Optional[str] = None
    secret: typing.Optional[str] = None


class PassportElementError(BaseModel):
    pass


class PassportElementErrorDataField(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    field_name: typing.Optional[str] = None
    data_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class PassportElementErrorFrontSide(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class PassportElementErrorReverseSide(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class PassportElementErrorSelfie(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class PassportElementErrorFile(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class PassportElementErrorFiles(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hashes: typing.Optional[typing.List[str]] = None
    message: typing.Optional[str] = None


class PassportElementErrorTranslationFile(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class PassportElementErrorTranslationFiles(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    file_hashes: typing.Optional[typing.List[str]] = None
    message: typing.Optional[str] = None


class PassportElementErrorUnspecified(BaseModel):
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    element_hash: typing.Optional[str] = None
    message: typing.Optional[str] = None


class Game(BaseModel):
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None
    photo: typing.Optional[typing.List['PhotoSize']] = None
    text: typing.Optional[str] = None
    text_entities: typing.Optional[typing.List['MessageEntity']] = None
    animation: typing.Optional['Animation'] = None


class CallbackGame(BaseModel):
    pass


class GameHighScore(BaseModel):
    position: typing.Optional[int] = None
    user: typing.Optional['User'] = None
    score: typing.Optional[int] = None


for v in locals().copy().values():
    if inspect.isclass(v) and issubclass(v, BaseModel):
        v.update_forward_refs()

__all__ = (
    'Error',
    'Update',
    'WebhookInfo',
    'User',
    'Chat',
    'Message',
    'MessageId',
    'MessageEntity',
    'PhotoSize',
    'Animation',
    'Audio',
    'Document',
    'Video',
    'VideoNote',
    'Voice',
    'Contact',
    'Dice',
    'PollOption',
    'PollAnswer',
    'Poll',
    'Location',
    'Venue',
    'WebAppData',
    'ProximityAlertTriggered',
    'MessageAutoDeleteTimerChanged',
    'VideoChatScheduled',
    'VideoChatStarted',
    'VideoChatEnded',
    'VideoChatParticipantsInvited',
    'UserProfilePhotos',
    'File',
    'WebAppInfo',
    'ReplyKeyboardMarkup',
    'KeyboardButton',
    'KeyboardButtonPollType',
    'ReplyKeyboardRemove',
    'InlineKeyboardMarkup',
    'InlineKeyboardButton',
    'LoginUrl',
    'CallbackQuery',
    'ForceReply',
    'ChatPhoto',
    'ChatInviteLink',
    'ChatAdministratorRights',
    'ChatMember',
    'ChatMemberOwner',
    'ChatMemberAdministrator',
    'ChatMemberMember',
    'ChatMemberRestricted',
    'ChatMemberLeft',
    'ChatMemberBanned',
    'ChatMemberUpdated',
    'ChatJoinRequest',
    'ChatPermissions',
    'ChatLocation',
    'BotCommand',
    'BotCommandScope',
    'BotCommandScopeDefault',
    'BotCommandScopeAllPrivateChats',
    'BotCommandScopeAllGroupChats',
    'BotCommandScopeAllChatAdministrators',
    'BotCommandScopeChat',
    'BotCommandScopeChatAdministrators',
    'BotCommandScopeChatMember',
    'MenuButton',
    'MenuButtonCommands',
    'MenuButtonWebApp',
    'MenuButtonDefault',
    'ResponseParameters',
    'InputMedia',
    'InputMediaPhoto',
    'InputMediaVideo',
    'InputMediaAnimation',
    'InputMediaAudio',
    'InputMediaDocument',
    'InputFile',
    'Sticker',
    'StickerSet',
    'MaskPosition',
    'InlineQuery',
    'InlineQueryResult',
    'InlineQueryResultArticle',
    'InlineQueryResultPhoto',
    'InlineQueryResultGif',
    'InlineQueryResultMpeg4Gif',
    'InlineQueryResultVideo',
    'InlineQueryResultAudio',
    'InlineQueryResultVoice',
    'InlineQueryResultDocument',
    'InlineQueryResultLocation',
    'InlineQueryResultVenue',
    'InlineQueryResultContact',
    'InlineQueryResultGame',
    'InlineQueryResultCachedPhoto',
    'InlineQueryResultCachedGif',
    'InlineQueryResultCachedMpeg4Gif',
    'InlineQueryResultCachedSticker',
    'InlineQueryResultCachedDocument',
    'InlineQueryResultCachedVideo',
    'InlineQueryResultCachedVoice',
    'InlineQueryResultCachedAudio',
    'InputMessageContent',
    'InputTextMessageContent',
    'InputLocationMessageContent',
    'InputVenueMessageContent',
    'InputContactMessageContent',
    'InputInvoiceMessageContent',
    'ChosenInlineResult',
    'SentWebAppMessage',
    'LabeledPrice',
    'Invoice',
    'ShippingAddress',
    'OrderInfo',
    'ShippingOption',
    'SuccessfulPayment',
    'ShippingQuery',
    'PreCheckoutQuery',
    'PassportData',
    'PassportFile',
    'EncryptedPassportElement',
    'EncryptedCredentials',
    'PassportElementError',
    'PassportElementErrorDataField',
    'PassportElementErrorFrontSide',
    'PassportElementErrorReverseSide',
    'PassportElementErrorSelfie',
    'PassportElementErrorFile',
    'PassportElementErrorFiles',
    'PassportElementErrorTranslationFile',
    'PassportElementErrorTranslationFiles',
    'PassportElementErrorUnspecified',
    'Game',
    'CallbackGame',
    'GameHighScore',
)



import typing
from abc import ABC, abstractmethod


class ABCPolling(ABC):
    @abstractmethod
    async def get_updates(self) -> typing.Any:
        pass

    @abstractmethod
    async def listen(self) -> typing.AsyncIterator[dict]:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass



import logging
import typing

from choicelib import choice_in_order
from typing_extensions import Protocol


class JSONModule(Protocol):
    def loads(self, s: str) -> typing.Union[dict, list]:
        ...

    def dumps(self, o: typing.Union[dict, list]) -> str:
        ...


json: JSONModule = choice_in_order(
    ["ujson", "hyperjson", "orjson"], do_import=True, default="json"
)
logger = choice_in_order(["loguru"], do_import=True, default="logging")

if logger.__name__ == "logging":
    logger = logging.getLogger("telegrinder")
elif logger.__name__ == "loguru":
    logger = getattr(logger, "logger")



from abc import ABC, abstractmethod
import typing

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch import Dispatch
    from telegrinder.api import API


class ABCScenario(ABC):
    @abstractmethod
    def wait(self, api: "API", dispatch: "Dispatch") -> typing.Any:
        pass



from typing import Optional


class Button:
    def __init__(
        self,
        text: str,
        request_contact: Optional[bool] = False,
        request_location: Optional[bool] = False,
        request_poll: Optional[dict] = None,
    ):
        self.text = text
        self.request_contact = request_contact
        self.request_location = request_location
        self.request_poll = request_poll

    def get_data(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}


class InlineButton:
    def __init__(
        self,
        text: str,
        url: str = None,
        login_url: dict = None,
        pay: bool = None,
        callback_data: str = None,
        callback_game: dict = None,
        switch_inline_query: str = None,
        switch_inline_query_current_chat: str = None,
    ):
        self.text = text
        self.url = url
        self.login_url = login_url
        self.pay = pay
        self.callback_data = callback_data
        self.callback_game = callback_game
        self.switch_inline_query = switch_inline_query
        self.switch_inline_query_current_chat = switch_inline_query_current_chat

    def get_data(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}



import typing
from telegrinder.http_.abc import ABCClient
from aiohttp import ClientSession, TCPConnector
from telegrinder.modules import json, JSONModule

if typing.TYPE_CHECKING:
    from aiohttp import ClientResponse


class AiohttpClient(ABCClient):
    def __init__(
        self,
        loop: typing.Optional[ClientSession] = None,
        session: typing.Optional[ClientSession] = None,
        json_processing_module: typing.Optional[JSONModule] = None,
        **session_params
    ):
        self.loop = loop
        self.session = session
        self.json_processing_module = json_processing_module or json
        self.session_params = session_params

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> "ClientResponse":
        if not self.session:
            self.session = ClientSession(
                connector=TCPConnector(ssl=False),
                json_serialize=self.json_processing_module.dumps,
                **self.session_params,
            )
        async with self.session.request(
            url=url, method=method, data=data, **kwargs
        ) as response:
            await response.read()
            return response

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> dict:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="utf-8", loads=self.json_processing_module.loads, content_type=None
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.text(encoding="utf-8")

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: typing.Optional[dict] = None,
        **kwargs
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    def __del__(self):
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector.close()
            self.session._connector = None
